@@ -1,3 +1,5 @@
+{-# LANGUAGE CPP #-}
+#define dummy	-- just to ensure cpp gets called on this file
 -- | This is just a convenient way of bunching the XML combinators
 --   together with some other things you are likely to want at the
 --   same time.
@@ -29,6 +31,7 @@ import Text.XML.HaXml.Escape
 
 import Text.PrettyPrint.HughesPJ  (render)
 
--- | The version of the library (currently "1.13.2").
+-- | The version of the library.
 version :: String
-version  = "1.13.2"
+version  = MYVERSION
+		-- expect cpp to fill in value
