@@ -20,7 +20,8 @@ import Monad
 
 import Text.XML.HaXml.Types
 import Text.XML.HaXml.Lex
-import Text.ParserCombinators.HuttonMeijerWallace
+import Text.XML.HaXml.Posn
+import Text.ParserCombinators.Poly.Plain
 
 --  #define DEBUG
  
@@ -37,45 +38,47 @@ debug :: Monad m => String -> m ()
 debug s = trace s (return ())
 #else
 debug :: Monad m => String -> m ()
-debug s = return ()
+debug _ = return ()
 #endif
 
 
 -- | The first argument is the name of the file, the second is the string
 --   contents of the file.  The result is the generic representation of
---   an XML document.
-htmlParse :: String -> String -> Document
-htmlParse name = simplify . sanitycheck . Prelude.either error id
-                 . papply' document () . xmlLex name
-
-sanitycheck :: Show p => [(a,s,[Either String (p,t)])] -> a
-sanitycheck [] = error "***Error at line 0: document not HTML?"
-sanitycheck ((x,_,[]):_) = x
-sanitycheck ((x,_,s@(Right (n,_):_)):xs) =
-  error ("***Error at "++show n++": data beyond end of parsed document")
+--   an XML document.  Any errors cause program failure with message to stderr.
+htmlParse :: String -> String -> Document Posn
+htmlParse name = Prelude.either error id . htmlParse' name
 
+-- | The first argument is the name of the file, the second is the string
+--   contents of the file.  The result is the generic representation of
+--   an XML document.  Any parsing errors are returned in the @Either@ type.
+htmlParse' :: String -> String -> Either String (Document Posn)
+htmlParse' name = Prelude.either Left (Right . simplify) . fst
+                  . runParser document . xmlLex name
 
 ---- Document simplification ----
 
-simplify :: Document -> Document
+simplify :: Document i -> Document i
 simplify (Document p st (Elem n avs cs) ms) =
     Document p st (Elem n avs (deepfilter simp cs)) ms
   where
-    simp (CElem (Elem "null" [] [])) = False
-    simp (CElem (Elem  n     _  [])) | n `elem` ["font","p","i","b","em"
-                                                ,"tt","big","small"] = False
-    simp (CString False s) | all isSpace s = False
+    simp (CElem (Elem "null" [] []) _) = False
+    simp (CElem (Elem  t     _  []) _) | t `elem` ["font","p","i","b","em"
+                                                  ,"tt","big","small"] = False
+ -- simp (CString False s _) | all isSpace s = False
     simp _ = True
-    deepfilter p =
-        filter p . map (\c-> case c of
-                          (CElem (Elem n avs cs)) -> CElem (Elem n avs (deepfilter p cs))
-                          _ -> c)
+    deepfilter f =
+        filter f . map (\c-> case c of
+                          CElem (Elem t avs cs) i
+                                  -> CElem (Elem t avs (deepfilter f cs)) i
+                          _       -> c)
 
 -- opening any of these, they close again immediately
+selfclosingtags :: [String]
 selfclosingtags = ["img","hr","br","meta","col","link","base"
                   ,"param","area","frame","input"]
 
 --closing this, implicitly closes any of those which are contained in it
+closeInnerTags :: [(String,[String])]
 closeInnerTags =
   [ ("ul",      ["li"])
   , ("ol",      ["li"])
@@ -119,83 +122,102 @@ t `closes` "p"
 _ `closes` _ = False
 
 
-
----- Misc ----
-
-fst3 (a,_,_) = a
-snd3 (_,a,_) = a
-thd3 (_,_,a) = a
-
-
-
 ---- Auxiliary Parsing Functions ----
 
-type HParser a = Parser () (Posn,TokenT) String a
+type HParser a = Parser (Posn,TokenT) a
+
+tok :: TokenT -> HParser TokenT
+tok t = do (p,t') <- next
+           case t' of TokError _    -> report failBad (show t) p t'
+                      _ | t'==t     -> return t
+                        | otherwise -> report fail (show t) p t'
 
 name :: HParser Name
-name = do {(p,TokName s) <- item; return s}
+--name = do {(p,TokName s) <- next; return s}
+name = do (p,tok) <- next
+          case tok of 
+            TokName s  -> return s 
+            TokError _ -> report failBad "a name" p tok
+            _          -> report fail "a name" p tok
 
 string, freetext :: HParser String
-string   = do {(p,TokName s) <- item; return s}
-freetext = do {(p,TokFreeText s) <- item; return s}
+string   = do (p,t) <- next
+              case t of TokName s -> return s
+                        _         -> report fail "text" p t
+freetext = do (p,t) <- next
+              case t of TokFreeText s -> return s
+                        _             -> report fail "text" p t
 
 maybe :: HParser a -> HParser (Maybe a)
 maybe p =
-    ( p >>= return . Just) +++
+    ( p >>= return . Just) `onFail`
     ( return Nothing)
 
 either :: HParser a -> HParser b -> HParser (Either a b)
 either p q =
-    ( p >>= return . Left) +++
+    ( p >>= return . Left) `onFail`
     ( q >>= return . Right)
 
 word :: String -> HParser ()
-word s = P (\st inp-> case inp of {
-               (Left err: _) -> Left err;
-               (Right (p,TokName n):ts) -> if s==n then Right [((),st,ts)]
-                                           else Right [];
-               (Right (p,TokFreeText n):ts) -> if s==n then Right [((),st,ts)]
-                                               else Right [];
-               ts -> Right [] } )
+word s = do { x <- next
+            ; case x of
+                (_p,TokName n)     | s==n -> return ()
+                (_p,TokFreeText n) | s==n -> return ()
+                ( p,t@(TokError _)) -> report failBad (show s) p t
+                ( p,t) -> report fail (show s) p t
+            }
 
 posn :: HParser Posn
-posn = P (\st inp-> case inp of {
-                     (Left err: _)   -> Left err;
-                     (Right (p,_):_) -> Right [(p,st,inp)];
-                     [] -> Right [(Pn "unknown" 0 0 Nothing,st,inp)]; } )
+posn = do { x@(p,_) <- next
+          ; reparse [x]
+          ; return p
+          } `onFail` return noPos
 
 nmtoken :: HParser NmToken
-nmtoken = (string +++ freetext)
+nmtoken = (string `onFail` freetext)
 
+failP, failBadP :: String -> HParser a
+failP msg    = do { p <- posn; fail (msg++"\n    at "++show p) }
+failBadP msg = do { p <- posn; failBad (msg++"\n    at "++show p) }
+
+report :: (String->HParser a) -> String -> Posn -> TokenT -> HParser a
+report fail expect p t = fail ("Expected "++show expect++" but found "++show t
+                               ++"\n  at "++show p)
+
+adjustErrP :: HParser a -> (String->String) -> HParser a
+p `adjustErrP` f = p `onFail` do pn <- posn
+                                 (p `adjustErr` f) `adjustErr` (++show pn)
 
 ---- XML Parsing Functions ----
 
-document :: HParser Document
+document :: HParser (Document Posn)
 document = do
-    p     <- prolog `elserror` "unrecognisable XML prolog"
-    es    <- many1 (element "HTML document")
+    p     <- prolog `adjustErr` ("unrecognisable XML prolog\n"++)
+    ht    <- many1 (element "HTML document")
     ms    <- many misc
-    return (Document p emptyST (case map snd es of
+    return (Document p emptyST (case map snd ht of
                                   [e] -> e
-                                  es  -> Elem "html" [] (map CElem es))
+                                  es  -> Elem "html" [] (map mkCElem es))
                                ms)
+  where mkCElem e = CElem e noPos
 
 comment :: HParser Comment
 comment = do
-    bracket (tok TokCommentOpen) freetext (tok TokCommentClose)
+    bracket (tok TokCommentOpen) (tok TokCommentClose) freetext
 
 processinginstruction :: HParser ProcessingInstruction
 processinginstruction = do
     tok TokPIOpen
-    n <- string  `elserror` "processing instruction has no target"
-    f <- freetext
-    (tok TokPIClose +++ tok TokAnyClose) `elserror` "missing ?> or >"
-    return (n, f)
+    commit $ do
+      n <- string  `onFail` failP "processing instruction has no target"
+      f <- freetext
+      (tok TokPIClose `onFail` tok TokAnyClose) `onFail` failP "missing ?> or >"
+      return (n, f)
 
 cdsect :: HParser CDSect
 cdsect = do
     tok TokSectionOpen
-    bracket (tok (TokSection CDATAx)) chardata (tok TokSectionClose)
+    bracket (tok (TokSection CDATAx)) (tok TokSectionClose) chardata
 
 prolog :: HParser Prolog
 prolog = do
@@ -208,30 +230,29 @@ prolog = do
 xmldecl :: HParser XMLDecl
 xmldecl = do
     tok TokPIOpen
-    (word "xml" +++ word "XML")
+    (word "xml" `onFail` word "XML")
     p <- posn
     s <- freetext
-    tok TokPIClose `elserror` "missing ?> in <?xml ...?>"
-    (raise . papply' aux () . xmlReLex p) s
+    tok TokPIClose `onFail` failBadP "missing ?> in <?xml ...?>"
+    (Prelude.either failP return . fst . runParser aux . xmlReLex p) s
   where
     aux = do
-      v <- versioninfo  `elserror` "missing XML version info"
+      v <- versioninfo  `onFail` failP "missing XML version info"
       e <- maybe encodingdecl
       s <- maybe sddecl
       return (XMLDecl v e s)
-    raise (Left err) = mzero `elserror` err
-    raise (Right ok) = (return . fst3 . head) ok
 
 versioninfo :: HParser VersionInfo
 versioninfo = do
-    (word "version" +++ word "VERSION")
+    (word "version" `onFail` word "VERSION")
     tok TokEqual
-    bracket (tok TokQuote) freetext (tok TokQuote)
+    bracket (tok TokQuote) (tok TokQuote) freetext
 
 misc :: HParser Misc
 misc = 
-    ( comment >>= return . Comment) +++
-    ( processinginstruction >>= return . PI)
+    oneOf' [ ("<!--comment-->", comment >>= return . Comment)
+           , ("<?PI?>",         processinginstruction >>= return . PI)
+           ]
 
 
 -- Question: for HTML, should we disallow in-line DTDs, allowing only externals?
@@ -241,22 +262,21 @@ doctypedecl :: HParser DocTypeDecl
 doctypedecl = do
     tok TokSpecialOpen
     tok (TokSpecial DOCTYPEx)
-    n <- name
-    eid <- maybe externalid
---  es <- maybe (bracket (tok TokSqOpen)
---                       (many markupdecl)
---                       (tok TokSqClose))
-    tok TokAnyClose  `elserror` "missing > in DOCTYPE decl"
---  return (DTD n eid (case es of { Nothing -> []; Just e -> e }))
-    return (DTD n eid [])
+    commit $ do
+      n <- name
+      eid <- maybe externalid
+--    es <- maybe (bracket (tok TokSqOpen) (tok TokSqClose)) (many markupdecl)
+      tok TokAnyClose  `onFail` failP "missing > in DOCTYPE decl"
+--    return (DTD n eid (case es of { Nothing -> []; Just e -> e }))
+      return (DTD n eid [])
 
 --markupdecl :: HParser MarkupDecl
 --markupdecl =
---    ( elementdecl >>= return . Element) +++
---    ( attlistdecl >>= return . AttList) +++
---    ( entitydecl >>= return . Entity) +++
---    ( notationdecl >>= return . Notation) +++
---    ( misc >>= return . MarkupMisc) +++
+--    ( elementdecl >>= return . Element) `onFail`
+--    ( attlistdecl >>= return . AttList) `onFail`
+--    ( entitydecl >>= return . Entity) `onFail`
+--    ( notationdecl >>= return . Notation) `onFail`
+--    ( misc >>= return . MarkupMisc) `onFail`
 --    PEREF(MarkupPE,markupdecl)
 --
 --extsubset :: HParser ExtSubset
@@ -267,19 +287,19 @@ doctypedecl = do
 --
 --extsubsetdecl :: HParser ExtSubsetDecl
 --extsubsetdecl =
---    ( markupdecl >>= return . ExtMarkupDecl) +++
---    ( conditionalsect >>= return . ExtConditionalSect) +++
+--    ( markupdecl >>= return . ExtMarkupDecl) `onFail`
+--    ( conditionalsect >>= return . ExtConditionalSect) `onFail`
 --    PEREF(ExtPEReference,extsubsetdecl)
 
 sddecl :: HParser SDDecl
 sddecl = do
-    (word "standalone" +++ word "STANDALONE")
-    tok TokEqual `elserror` "missing = in 'standalone' decl"
-    bracket (tok TokQuote)
-            ( (word "yes" >> return True) +++
-              (word "no" >> return False) `elserror`
-              "'standalone' decl requires 'yes' or 'no' value" )
-            (tok TokQuote)
+    (word "standalone" `onFail` word "STANDALONE")
+    commit $ do
+      tok TokEqual `onFail` failP "missing = in 'standalone' decl"
+      bracket (tok TokQuote) (tok TokQuote)
+              ( (word "yes" >> return True) `onFail`
+                (word "no" >> return False) `onFail`
+                failP "'standalone' decl requires 'yes' or 'no' value" )
 
 
 
@@ -290,7 +310,7 @@ sddecl = do
 -- earliest opportunity.
 type Stack = [(Name,[Attribute])]
 
-element :: Name -> HParser (Stack,Element)
+element :: Name -> HParser (Stack,Element Posn)
 element ctx =
   do
     tok TokAnyOpen
@@ -305,29 +325,32 @@ element ctx =
          -- complete the parse straightaway.
          ( do tok TokEndClose	-- self-closing <tag /> 
               debug (e++"[+]")
-              return ([], Elem e avs [])) +++
+              return ([], Elem e avs [])) `onFail`
      --  ( do tok TokAnyClose	-- sequence <tag></tag>	(**not HTML?**)
      --       debug (e++"[+")
-     --       n <- bracket (tok TokEndOpen) name (tok TokAnyClose)
+     --       n <- bracket (tok TokEndOpen) (tok TokAnyClose) name
      --       debug "]"
      --       if e == (map toLower n :: Name) 
      --         then return ([], Elem e avs [])      
-     --         else return (error "no nesting in empty tag")) +++
+     --         else return (error "no nesting in empty tag")) `onFail`
          ( do tok TokAnyClose	-- <tag> with no close (e.g. <IMG>)
               debug (e++"[+]")
               return ([], Elem e avs []))
       else
         (( do tok TokEndClose
               debug (e++"[]")
-              return ([], Elem e avs [])) +++
-         ( do tok TokAnyClose `elserror` "missing > or /> in element tag"
+              return ([], Elem e avs [])) `onFail`
+         ( do tok TokAnyClose `onFail` failP "missing > or /> in element tag"
               debug (e++"[")
-              zz <- many (content e)
-           -- (if null zz then return (error ("empty content in context: "++e)) else return ())
+           -- zz <- many (content e)
+           -- n <- bracket (tok TokEndOpen) (tok TokAnyClose) name
+              zz <- manyFinally (content e)
+                                (tok TokEndOpen)
+              n <- name
+              commit (tok TokAnyClose)
+              debug "]"
               let (ss,cs) = unzip zz
               let s       = if null ss then [] else last ss
-              n <- bracket (tok TokEndOpen) name (tok TokAnyClose)
-              debug "]"
               ( if e == (map toLower n :: Name) then
                   do unparse (reformatTags (closeInner e s))
                      debug "^"
@@ -336,7 +359,7 @@ element ctx =
                   do unparse [TokEndOpen, TokName n, TokAnyClose]
                      debug "-"
                      return (((e,avs):s), Elem e avs cs))
-         ) `elserror` ("failed to repair non-matching tags in context: "++ctx)))
+         ) `onFail` failP ("failed to repair non-matching tags in context: "++ctx)))
 
 closeInner :: Name -> [(Name,[Attribute])] -> [(Name,[Attribute])]
 closeInner c ts =
@@ -344,27 +367,33 @@ closeInner c ts =
       (Just these) -> filter ((`notElem` these).fst) ts
       Nothing      -> ts
 
+unparse :: [TokenT] -> Parser (Posn, TokenT) ()
 unparse ts = do p <- posn
-                reparse (map Right (zip (repeat p) ts))
+                reparse (zip (repeat p) ts)
 
+reformatAttrs :: [(String, AttValue)] -> [TokenT]
 reformatAttrs avs = concatMap f0 avs
-    where f0 (a, AttValue [Left s]) = [TokName a, TokEqual, TokQuote,
-                                       TokFreeText s, TokQuote]
+    where f0 (a, v@(AttValue _)) = [TokName a, TokEqual, TokQuote,
+                                       TokFreeText (show v), TokQuote]
+
+reformatTags :: [(String, [(String, AttValue)])] -> [TokenT]
 reformatTags ts = concatMap f0 ts
     where f0 (t,avs) = [TokAnyOpen, TokName t]++reformatAttrs avs++[TokAnyClose]
 
-content :: Name -> HParser (Stack,Content)
-content ctx =
-    ( element ctx >>= \(s,e)-> return (s, CElem e)) +++
-    ( chardata >>= \s-> return ([], CString False s)) +++
-    ( reference >>= \r-> return ([], CRef r)) +++
-    ( cdsect >>= \c-> return ([], CString True c)) +++
-    ( misc >>= \m->  return ([], CMisc m))
+content :: Name -> HParser (Stack,Content Posn)
+content ctx = do { p <- posn ; content' p }
+  where content' p = oneOf'
+          [ ( "element", element ctx >>= \(s,e)-> return (s, CElem e p))
+          , ( "chardata", chardata >>= \s-> return ([], CString False s p))
+          , ( "reference", reference >>= \r-> return ([], CRef r p))
+          , ( "cdsect", cdsect >>= \c-> return ([], CString True c p))
+          , ( "misc", misc >>= \m->  return ([], CMisc m p))
+          ] `adjustErrP` ("when looking for a content item,\n"++)
 
 ----
 elemtag :: HParser ElemTag
 elemtag = do
-    n <- name `elserror` "malformed element tag"
+    n <- name `adjustErrBad` ("malformed element tag\n"++)
     as <- many attribute
     return (ElemTag (map toLower n) as)
 
@@ -372,7 +401,7 @@ attribute :: HParser Attribute
 attribute = do
     n <- name
     v <- (do tok TokEqual
-             attvalue) +++
+             attvalue) `onFail`
          (return (AttValue [Left "TRUE"]))
     return (map toLower n,v)
 
@@ -380,49 +409,47 @@ attribute = do
 --elementdecl = do
 --    tok TokSpecialOpen
 --    tok (TokSpecial ELEMENTx)
---    n <- name `elserror` "missing identifier in ELEMENT decl"
---    c <- contentspec `elserror` "missing content spec in ELEMENT decl"
---    tok TokAnyClose `elserror` "expected > terminating ELEMENT decl"
+--    n <- name `onFail` failP "missing identifier in ELEMENT decl"
+--    c <- contentspec `onFail` failP "missing content spec in ELEMENT decl"
+--    tok TokAnyClose `onFail` failP "expected > terminating ELEMENT decl"
 --    return (ElementDecl n c)
 --
 --contentspec :: HParser ContentSpec
 --contentspec =
---    ( word "EMPTY" >> return EMPTY) +++
---    ( word "ANY" >> return ANY) +++
---    ( mixed >>= return . Mixed) +++
---    ( cp >>= return . ContentSpec) +++
+--    ( word "EMPTY" >> return EMPTY) `onFail`
+--    ( word "ANY" >> return ANY) `onFail`
+--    ( mixed >>= return . Mixed) `onFail`
+--    ( cp >>= return . ContentSpec) `onFail`
 --    PEREF(ContentPE,contentspec)
 --
 --choice :: HParser [CP]
 --choice = do
---    bracket (tok TokBraOpen)
+--    bracket (tok TokBraOpen) (tok TokBraClose)
 --            (cp `sepby1` (tok TokPipe))
---            (tok TokBraClose)
 --
 --sequence :: HParser [CP]
 --sequence = do
---    bracket (tok TokBraOpen)
+--    bracket (tok TokBraOpen) (tok TokBraClose)
 --            (cp `sepby1` (tok TokComma))
---            (tok TokBraClose)
 --
 --cp :: HParser CP
 --cp =
 --    ( do n <- name
 --         m <- modifier
---         return (TagName n m)) +++
+--         return (TagName n m)) `onFail`
 --    ( do ss <- sequence
 --         m <- modifier
---         return (Seq ss m)) +++
+--         return (Seq ss m)) `onFail`
 --    ( do cs <- choice
 --         m <- modifier
---         return (Choice cs m)) +++
+--         return (Choice cs m)) `onFail`
 --    PEREF(CPPE,cp)
 --
 --modifier :: HParser Modifier
 --modifier =
---    ( tok TokStar >> return Star) +++
---    ( tok TokQuery >> return Query) +++
---    ( tok TokPlus >> return Plus) +++
+--    ( tok TokStar >> return Star) `onFail`
+--    ( tok TokQuery >> return Query) `onFail`
+--    ( tok TokPlus >> return Plus) `onFail`
 --    ( return None)
 --
 --mixed :: HParser Mixed
@@ -432,7 +459,7 @@ attribute = do
 --    word "PCDATA"
 --    cont
 --  where
---    cont = ( tok TokBraClose >> return PCDATA) +++
+--    cont = ( tok TokBraClose >> return PCDATA) `onFail`
 --           ( do cs <- many ( do tok TokPipe
 --                                n <- name
 --                                return n)
@@ -444,56 +471,54 @@ attribute = do
 --attlistdecl = do
 --    tok TokSpecialOpen
 --    tok (TokSpecial ATTLISTx)
---    n <- name `elserror` "missing identifier in ATTLIST"
+--    n <- name `onFail` failP "missing identifier in ATTLIST"
 --    ds <- many attdef
---    tok TokAnyClose `elserror` "missing > terminating ATTLIST"
+--    tok TokAnyClose `onFail` failP "missing > terminating ATTLIST"
 --    return (AttListDecl n ds)
 --
 --attdef :: HParser AttDef
 --attdef = do
 --    n <- name
---    t <- atttype `elserror` "missing attribute type in attlist defn"
+--    t <- atttype `onFail` failP "missing attribute type in attlist defn"
 --    d <- defaultdecl
 --    return (AttDef n t d)
 --
 --atttype :: HParser AttType
 --atttype =
---    ( word "CDATA" >> return StringType) +++
---    ( tokenizedtype >>= return . TokenizedType) +++
+--    ( word "CDATA" >> return StringType) `onFail`
+--    ( tokenizedtype >>= return . TokenizedType) `onFail`
 --    ( enumeratedtype >>= return . EnumeratedType)
 --
 --tokenizedtype :: HParser TokenizedType
 --tokenizedtype =
---    ( word "ID" >> return ID) +++
---    ( word "IDREF" >> return IDREF) +++
---    ( word "IDREFS" >> return IDREFS) +++
---    ( word "ENTITY" >> return ENTITY) +++
---    ( word "ENTITIES" >> return ENTITIES) +++
---    ( word "NMTOKEN" >> return NMTOKEN) +++
+--    ( word "ID" >> return ID) `onFail`
+--    ( word "IDREF" >> return IDREF) `onFail`
+--    ( word "IDREFS" >> return IDREFS) `onFail`
+--    ( word "ENTITY" >> return ENTITY) `onFail`
+--    ( word "ENTITIES" >> return ENTITIES) `onFail`
+--    ( word "NMTOKEN" >> return NMTOKEN) `onFail`
 --    ( word "NMTOKENS" >> return NMTOKENS)
 --
 --enumeratedtype :: HParser EnumeratedType
 --enumeratedtype =
---    ( notationtype >>= return . NotationType) +++
+--    ( notationtype >>= return . NotationType) `onFail`
 --    ( enumeration >>= return . Enumeration)
 --
 --notationtype :: HParser NotationType
 --notationtype = do
 --    word "NOTATION"
---    bracket (tok TokBraOpen)
+--    bracket (tok TokBraOpen) (tok TokBraClose)
 --            (name `sepby1` (tok TokPipe))
---            (tok TokBraClose)
 --
 --enumeration :: HParser Enumeration
 --enumeration =
---    bracket (tok TokBraOpen)
+--    bracket (tok TokBraOpen) (tok TokBraClose)
 --            (nmtoken `sepby1` (tok TokPipe))
---            (tok TokBraClose)
 --
 --defaultdecl :: HParser DefaultDecl
 --defaultdecl =
---    ( tok TokHash >> word "REQUIRED" >> return REQUIRED) +++
---    ( tok TokHash >> word "IMPLIED" >> return IMPLIED) +++
+--    ( tok TokHash >> word "REQUIRED" >> return REQUIRED) `onFail`
+--    ( tok TokHash >> word "IMPLIED" >> return IMPLIED) `onFail`
 --    ( do f <- maybe (tok TokHash >> word "FIXED" >> return FIXED)
 --         a <- attvalue
 --         return (DefaultTo a f))
@@ -502,15 +527,15 @@ attribute = do
 --conditionalsect =
 --    ( do tok TokSectionOpen
 --         tok (TokSection INCLUDEx)
---         tok TokSqOpen `elserror` "missing [ after INCLUDE"
---         i <- extsubsetdecl `elserror` "missing ExtSubsetDecl in INCLUDE"
---         tok TokSectionClose `elserror` "missing ] after INCLUDE"
---         return (IncludeSect i)) +++
+--         tok TokSqOpen `onFail` failP "missing [ after INCLUDE"
+--         i <- extsubsetdecl `onFail` failP "missing ExtSubsetDecl in INCLUDE"
+--         tok TokSectionClose `onFail` failP "missing ] after INCLUDE"
+--         return (IncludeSect i)) `onFail`
 --    ( do tok TokSectionOpen
 --         tok (TokSection IGNOREx)
---         tok TokSqOpen `elserror` "missing [ after IGNORE"
+--         tok TokSqOpen `onFail` failP "missing [ after IGNORE"
 --         i <- many ignoresectcontents
---         tok TokSectionClose `elserror` "missing ] after IGNORE"
+--         tok TokSectionClose `onFail` failP "missing ] after IGNORE"
 --         return (IgnoreSect i))
 --
 --ignoresectcontents :: HParser IgnoreSectContents
@@ -528,7 +553,7 @@ attribute = do
 
 reference :: HParser Reference
 reference = do
-    bracket (tok TokAmp) (freetext >>= val) (tok TokSemi)
+    bracket (tok TokAmp) (tok TokSemi) (freetext >>= val)
   where
     val ('#':'x':i) | all isHexDigit i
                     = return . RefChar . fst . head . readHex $ i
@@ -539,17 +564,17 @@ reference = do
 {-
 reference :: HParser Reference
 reference =
-    ( charref >>= return . RefChar) +++
+    ( charref >>= return . RefChar) `onFail`
     ( entityref >>= return . RefEntity)
 
 entityref :: HParser EntityRef
 entityref = do
-    n <- bracket (tok TokAmp) name (tok TokSemi)
+    n <- bracket (tok TokAmp) (tok TokSemi) name
     return n
 
 charref :: HParser CharRef
 charref = do
-    bracket (tok TokAmp) (freetext >>= readCharVal) (tok TokSemi)
+    bracket (tok TokAmp) (tok TokSemi) (freetext >>= readCharVal)
   where
     readCharVal ('#':'x':i) = return . fst . head . readHex $ i
     readCharVal ('#':i)     = return . fst . head . readDec $ i
@@ -558,11 +583,11 @@ charref = do
 
 --pereference :: HParser PEReference
 --pereference = do
---    bracket (tok TokPercent) nmtoken (tok TokSemi)
+--    bracket (tok TokPercent) (tok TokSemi) nmtoken
 --
 --entitydecl :: HParser EntityDecl
 --entitydecl =
---    ( gedecl >>= return . EntityGEDecl) +++
+--    ( gedecl >>= return . EntityGEDecl) `onFail`
 --    ( pedecl >>= return . EntityPEDecl)
 --
 --gedecl :: HParser GEDecl
@@ -570,8 +595,8 @@ charref = do
 --    tok TokSpecialOpen
 --    tok (TokSpecial ENTITYx)
 --    n <- name
---    e <- entitydef `elserror` "missing entity defn in G ENTITY decl"
---    tok TokAnyClose `elserror` "expected > terminating G ENTITY decl"
+--    e <- entitydef `onFail` failP "missing entity defn in G ENTITY decl"
+--    tok TokAnyClose `onFail` failP "expected > terminating G ENTITY decl"
 --    return (GEDecl n e)
 --
 --pedecl :: HParser PEDecl
@@ -580,30 +605,30 @@ charref = do
 --    tok (TokSpecial ENTITYx)
 --    tok TokPercent
 --    n <- name
---    e <- pedef `elserror` "missing entity defn in P ENTITY decl"
---    tok TokAnyClose `elserror` "expected > terminating P ENTITY decl"
+--    e <- pedef `onFail` failP "missing entity defn in P ENTITY decl"
+--    tok TokAnyClose `onFail` failP "expected > terminating P ENTITY decl"
 --    return (PEDecl n e)
 --
 --entitydef :: HParser EntityDef
 --entitydef =
---    ( entityvalue >>= return . DefEntityValue) +++
+--    ( entityvalue >>= return . DefEntityValue) `onFail`
 --    ( do eid <- externalid
 --         ndd <- maybe ndatadecl
 --         return (DefExternalID eid ndd))
 --
 --pedef :: HParser PEDef
 --pedef =
---    ( entityvalue >>= return . PEDefEntityValue) +++
+--    ( entityvalue >>= return . PEDefEntityValue) `onFail`
 --    ( externalid >>= return . PEDefExternalID)
 
 externalid :: HParser ExternalID
 externalid =
     ( do word "SYSTEM"
          s <- systemliteral
-         return (SYSTEM s)) +++
+         return (SYSTEM s)) `onFail`
     ( do word "PUBLIC"
          p <- pubidliteral
-         s <- (systemliteral +++ return (SystemLiteral ""))
+         s <- (systemliteral `onFail` return (SystemLiteral ""))
          return (PUBLIC p s))
 
 --ndatadecl :: HParser NDataDecl
@@ -612,14 +637,14 @@ externalid =
 --    n <- name
 --    return (NDATA n)
 
-textdecl :: HParser TextDecl
-textdecl = do
-    tok TokPIOpen
-    (word "xml" +++ word "XML")
-    v <- maybe versioninfo
-    e <- encodingdecl
-    tok TokPIClose `elserror` "expected ?> terminating text decl"
-    return (TextDecl v e)
+--textdecl :: HParser TextDecl
+--textdecl = do
+--    tok TokPIOpen
+--    (word "xml" `onFail` word "XML")
+--    v <- maybe versioninfo
+--    e <- encodingdecl
+--    tok TokPIClose `onFail` failP "expected ?> terminating text decl"
+--    return (TextDecl v e)
 
 --extparsedent :: HParser ExtParsedEnt
 --extparsedent = do
@@ -635,9 +660,9 @@ textdecl = do
 
 encodingdecl :: HParser EncodingDecl
 encodingdecl = do
-    (word "encoding" +++ word "ENCODING")
-    tok TokEqual `elserror` "expected = in 'encoding' decl"
-    f <- bracket (tok TokQuote) freetext (tok TokQuote)
+    (word "encoding" `onFail` word "ENCODING")
+    tok TokEqual `onFail` failBadP "expected = in 'encoding' decl"
+    f <- bracket (tok TokQuote) (tok TokQuote) freetext
     return (EncodingDecl f)
 
 --notationdecl :: HParser NotationDecl
@@ -646,46 +671,49 @@ encodingdecl = do
 --    word "NOTATION"
 --    n <- name
 --    e <- either externalid publicid
---    tok TokAnyClose `elserror` "expected > terminating NOTATION decl"
+--    tok TokAnyClose `onFail` failP "expected > terminating NOTATION decl"
 --    return (NOTATION n e)
 
-publicid :: HParser PublicID
-publicid = do
-    word "PUBLICID"
-    p <- pubidliteral
-    return (PUBLICID p)
+--publicid :: HParser PublicID
+--publicid = do
+--    word "PUBLICID"
+--    p <- pubidliteral
+--    return (PUBLICID p)
 
-entityvalue :: HParser EntityValue
-entityvalue = do
-    evs <- bracket (tok TokQuote) (many ev) (tok TokQuote)
-    return (EntityValue evs)
+--entityvalue :: HParser EntityValue
+--entityvalue = do
+--    evs <- bracket (tok TokQuote) (tok TokQuote) (many ev)
+--    return (EntityValue evs)
 
-ev :: HParser EV
-ev =
-    ( freetext >>= return . EVString) +++
---  PEREF(EVPERef,ev) +++
-    ( reference >>= return . EVRef)
+--ev :: HParser EV
+--ev =
+--    ( freetext >>= return . EVString) `onFail`
+-- -- PEREF(EVPERef,ev) `onFail`
+--    ( reference >>= return . EVRef)
 
 attvalue :: HParser AttValue
 attvalue =
-  ( do avs <- bracket (tok TokQuote)
+  ( do avs <- bracket (tok TokQuote) (tok TokQuote)
                       (many (either freetext reference))
-                      (tok TokQuote)
-       return (AttValue avs) ) +++
-  ( do p <- (tok TokPlus >> return ('+':))
-            +++ (tok TokHash >> return ('#':))
-            +++ (return id)
+       return (AttValue avs) ) `onFail`
+  ( do v <- nmtoken
+       s <- (tok TokPercent >> return "%") `onFail` return ""
+       return (AttValue [Left (v++s)]) ) `onFail`
+  ( do s <- oneOf [ tok TokPlus >> return "+"
+                  , tok TokHash >> return "#"
+                  ]
        v <- nmtoken
-       return (AttValue [Left (p v)]) )
+       return (AttValue [Left (s++v)]) ) `onFail`
+  failP "Badly formatted attribute value"
 
 systemliteral :: HParser SystemLiteral
 systemliteral = do
-    s <- bracket (tok TokQuote) freetext (tok TokQuote)
+    s <- bracket (tok TokQuote) (tok TokQuote) freetext
     return (SystemLiteral s)		-- note: need to fold &...; escapes
 
 pubidliteral :: HParser PubidLiteral
 pubidliteral = do
-    s <- bracket (tok TokQuote) freetext (tok TokQuote)
+    s <- bracket (tok TokQuote) (tok TokQuote) freetext
     return (PubidLiteral s)		-- note: need to fold &...; escapes
 
 chardata :: HParser CharData
